@@ -1,11 +1,11 @@
 package HTML::Template::Compiled::Parser;
-# $Id: Parser.pm,v 1.71 2007/10/09 18:24:59 tinita Exp $
+# $Id: Parser.pm 1102 2009-08-21 13:56:24Z tinita $
 use Carp qw(croak carp confess);
 use strict;
 use warnings;
 use base qw(Exporter);
 use HTML::Template::Compiled::Token qw(:tagtypes);
-our $VERSION = 0.07;
+our $VERSION = 0.11;
 my @vars;
 BEGIN {
 @vars = qw(
@@ -32,9 +32,10 @@ $UNTAINT                = 0;
 $DEFAULT_QUERY          = 0;
 $DEFAULT_TAGSTYLE       = [qw(classic comment asp)];
 
-use constant ATTR_TAGSTYLE => 0;
-use constant ATTR_TAGNAMES => 1;
-use constant ATTR_PERL     => 2;
+use constant ATTR_TAGSTYLE   => 0;
+use constant ATTR_TAGNAMES   => 1;
+use constant ATTR_PERL       => 2;
+use constant ATTR_EXPRESSION => 3;
 
 use constant T_VAR         => 'VAR';
 use constant T_IF          => 'IF';
@@ -71,6 +72,9 @@ sub get_tagnames { $_[0]->[ATTR_TAGNAMES] }
 sub set_perl   { $_[0]->[ATTR_PERL] = $_[1] }
 sub get_perl   { $_[0]->[ATTR_PERL] }
 
+sub set_expressions { $_[0]->[ATTR_EXPRESSION] = $_[1] }
+sub get_expressions { $_[0]->[ATTR_EXPRESSION] }
+
 sub add_tagnames {
     my ($self, $hash) = @_;
     my $open = $hash->{OPENING_TAG()};
@@ -79,6 +83,13 @@ sub add_tagnames {
     @{ $_[0]->[ATTR_TAGNAMES]->{CLOSING_TAG()} }{keys %$close} = values %$close;
 }
 
+sub remove_tags {
+    my ($self, @tags) = @_;
+    my $open = $self->[ATTR_TAGNAMES]->{OPENING_TAG()};
+    my $close = $self->[ATTR_TAGNAMES]->{CLOSING_TAG()};
+    delete @$open{@tags};
+    delete @$close{@tags};
+}
 
 my $_default_tags = {
     classic => ['<TMPL_'      ,'>',     '</TMPL_',      '>',  ],
@@ -100,28 +111,33 @@ sub default_tags {
     return $_default_tags;
 }
 
-my $default_validation = sub { exists $_[1]->{NAME} };
+my $default_validation = sub {
+    my ($p, $attr) = @_;
+    my $test = $p->get_expressions;
+    exists $attr->{NAME} or
+    ($p->get_expressions and exists $attr->{EXPR})
+};
 my %allowed_tagnames = (
     OPENING_TAG() => {
-        VAR         => [$default_validation, qw(NAME ESCAPE DEFAULT)],
+        VAR         => [$default_validation, qw(NAME ESCAPE DEFAULT EXPR)],
         # just an alias for VAR
-        '='         => [$default_validation, qw(NAME ESCAPE DEFAULT)],
-        IF_DEFINED  => [$default_validation, qw(NAME)],
-        IF          => [$default_validation, qw(NAME)],
-        UNLESS      => [$default_validation, qw(NAME)],
-        ELSIF       => [$default_validation, qw(NAME)],
+        '='         => [$default_validation, qw(NAME ESCAPE DEFAULT EXPR)],
+        IF_DEFINED  => [$default_validation, qw(NAME EXPR)],
+        IF          => [$default_validation, qw(NAME EXPR)],
+        UNLESS      => [$default_validation, qw(NAME EXPR)],
+        ELSIF       => [$default_validation, qw(NAME EXPR)],
         ELSE        => [undef, qw(NAME)],
-        WITH        => [$default_validation, qw(NAME)],
+        WITH        => [$default_validation, qw(NAME EXPR)],
         COMMENT     => [undef, qw(NAME)],
         VERBATIM    => [undef, qw(NAME)],
         NOPARSE     => [undef, qw(NAME)],
-        LOOP        => [$default_validation, qw(NAME ALIAS JOIN BREAK)],
-        WHILE       => [$default_validation, qw(NAME ALIAS BREAK)],
-        EACH        => [$default_validation, qw(NAME BREAK)],
-        SWITCH      => [$default_validation, qw(NAME)],
+        LOOP        => [$default_validation, qw(NAME ALIAS JOIN BREAK EXPR CONTEXT)],
+        WHILE       => [$default_validation, qw(NAME ALIAS BREAK EXPR)],
+        EACH        => [$default_validation, qw(NAME BREAK EXPR SORT REVERSE CONTEXT)],
+        SWITCH      => [$default_validation, qw(NAME EXPR)],
         CASE        => [undef, qw(NAME)],
-        INCLUDE_VAR => [$default_validation, qw(NAME)],
-        INCLUDE_STRING => [$default_validation, qw(NAME)],
+        INCLUDE_VAR => [$default_validation, qw(NAME EXPR)],
+        INCLUDE_STRING => [$default_validation, qw(NAME EXPR)],
         INCLUDE     => [$default_validation, qw(NAME)],
     },
     CLOSING_TAG() => {
@@ -146,6 +162,7 @@ sub init {
     my $tagnames = $args{tagnames} || {};
     my $tagstyle = $self->_create_tagstyle( $args{tagstyle} );
     $self->[ATTR_TAGSTYLE] = $tagstyle;
+    $self->[ATTR_EXPRESSION] = $args{use_expressions};
     $self->[ATTR_TAGNAMES] = {
         OPENING_TAG() => {
             %{ $allowed_tagnames{ OPENING_TAG() } },
@@ -253,7 +270,7 @@ sub find_attributes {
         $attr->{$key} = $val;
         $arg->{token} .= $orig;
     }
-    my $ok = $validate_sub ? $validate_sub->(undef, $attr) : 1;
+    my $ok = $validate_sub ? $validate_sub->($self, $attr) : 1;
     $self->_error_wrong_tag_syntax(
         $arg, $arg->{template}
     ) unless $ok;
@@ -264,7 +281,7 @@ sub find_attributes {
     my $callbacks_found_text;
     my $encode_tag = sub {
         my ( $p, $arg ) = @_;
-        HTML::Entities::encode_entities($arg->{token});
+        $arg->{token} = HTML::Template::Compiled::Utils::escape_html($arg->{token});
         $callbacks_found_text->[0]->($p, $arg);
         $arg->{token} = "";
     };
@@ -286,7 +303,7 @@ sub find_attributes {
     };
     my $default_callback_tag = sub {
         my ($self, $arg) = @_;
-        #print STDERR "####found tag ${$args{name}}\n";
+        #print STDERR "####found tag $arg->{name}, $arg->{open_or_close}\n";
         $arg->{line} += $arg->{token} =~ tr/\n//;
         my $class = 'HTML::Template::Compiled::Token::' .
             ($arg->{open_or_close} == OPENING_TAG
@@ -562,7 +579,7 @@ sub find_attribute {
         $orig .= $1;
     }
     #print STDERR "match '$$text' (?=$until|\\s)\n";
-    if ($arg->{template} =~ s{^ (\s* (['"]) ([^"]+) \2 \s*) }{}x) {
+    if ($arg->{template} =~ s{^ (\s* (['"]) (.+?) \2 \s*) }{}x) {
         #print STDERR qq{matched $2$3$2\n};
         $var = "$3";
         $orig .= $1;
@@ -587,7 +604,8 @@ sub find_attribute {
         T_CASE()   => [T_SWITCH],
     };
     $map[CLOSING_TAG] = {
-        IF         => [ T_IF, T_UNLESS, T_ELSE ],
+        IF         => [ T_IF, T_UNLESS, T_ELSE, T_IF_DEFINED ],
+        T_IF_DEFINED() => [ T_ELSE, T_IF_DEFINED ],
         UNLESS     => [T_UNLESS, T_ELSE, T_IF_DEFINED],
         ELSIF      => [ T_IF, T_UNLESS, T_IF_DEFINED ],
         LOOP       => [T_LOOP],
@@ -636,7 +654,7 @@ sub find_attribute {
         }
         elsif ($arg->{open_or_close} == CLOSING_TAG) {
             if (grep { $arg->{name} eq $_ } (
-                    T_IF, T_UNLESS, T_WITH, T_LOOP, T_WHILE, T_SWITCH
+                    T_IF, T_IF_DEFINED, T_UNLESS, T_WITH, T_LOOP, T_WHILE, T_SWITCH
                 )) {
                 pop @{ $arg->{stack} };
             }
